from .func_helpers import get_area
import numpy as np
from cxas.label_mapper import id2label_dict


def get_all_areas(mask, img=None, draw = False):
    out = {}
    for i in range(mask.shape[0]):
        x = get_area(mask[i].astype(np.uint8), 1, 1)
        out[id2label_dict[str(i)]+'_area'] = x
    return out 



import numpy as np
import math

def get_center_of_mask(mask):
    y_coords, x_coords = np.nonzero(mask)
    center = (np.mean(x_coords).astype(np.int32), np.mean(y_coords).astype(np.int32))
    return center

def get_centers(mask, ids):
    centers = [get_center_of_mask(mask[id][0]) for id in ids if mask[id][0].sum()>0]
    return centers

def get_area(mask, spacing_x, spacing_y):
    return (mask*spacing_x*spacing_y).sum()

def get_perimeter_from_contour(cnt, conversion_factor):
    segment_lengths = []
    for i in range(len(cnt)-1):
        x1, y1 = cnt[i][0]
        x2, y2 = cnt[i+1][0]
        segment_length = np.sqrt((x2-x1)**2 + (y2-y1)**2) * conversion_factor
        segment_lengths.append(segment_length)

    # Calculate total perimeter
    perimeter = sum(segment_lengths)
    return perimeter

def sort_by_distance(reference_point, coordinates):
    distances = []
    for point in coordinates:
        x1, y1 = reference_point
        x2, y2 = point
        dist = math.sqrt((x2 - x1)**2 + (y2 - y1)**2)
        distances.append((point, dist))
    distances.sort(key=lambda x: x[1])
    return [point for (point, dist) in distances]

def get_min_dist(points1, points2):
    out_points = []
    out_dists = []
    # import pdb; pdb.set_trace()
    for point in points1:
        differences = points2 - point
        distances = np.linalg.norm(differences, axis=1)
        min_index = np.argmin(distances)
        out_points += [points2[min_index]]
        out_dists += [distances[min_index]]
    # import pdb; pdb.set_trace()
    return out_points, out_dists


from .compactness import get_all_compactness
from .area import get_all_areas
from .centroid import get_centroids
from .cardiothoracic_ratio import get_cardiothoracic_ratio
from .spinecenter_distance import get_spine_center_distance
from .perimeter import get_all_perimeters
from .bounding_box import get_all_bounding_boxes
from .convexity import get_all_convexities
import numpy as np

class Extractor():
    def __init__(self,):
        """
        """
        self.methods = {
            'Cardio-Thoracic Ratio': get_cardiothoracic_ratio,
            'CTR': get_cardiothoracic_ratio,
            'Spine-Center Distance': get_spine_center_distance,
            'SCD': get_spine_center_distance,
            'perimeter': get_all_perimeters,
            'compactness': get_all_compactness,
            'area': get_all_areas,
            'centroid': get_centroids,
            'box': get_all_bounding_boxes,
            'convexity': get_all_convexities,

        }
    
    def extract(self, 
                file:np.array, 
                method:str, 
                image: np.array = None,
                draw:bool=False
               ) -> dict:
        """
        """
        assert method in list(self.methods.keys()), f'Method in question ({method}) is not yet implemented. Please write an issue if you want to have it implemented.'
        return self.methods[method](file, image, draw)


from PIL import Image
import SimpleITK as sitk
import numpy as np
import os
import json
import torch
import torch.nn.functional as F
from itertools import groupby
from cxas.label_mapper import id2label_dict, category_ids
from torch.utils.data import Dataset
import SimpleITK as sitk
import numpy as np
from cxas.io_utils.dicomseg_2d import write_dicom_seg
from cxas.io_utils.create_annotations import (
    get_coco_json_format,
    create_category_annotation,
)
from cxas.io_utils.mask_to_coco import toBox, mask_to_annotation
from pathlib import Path

this_directory = Path(__file__).parent

class FolderDataset(Dataset):
    """
    Dataset class to load images from a folder.
    """

    def __init__(self, path: str, gpus: str):
        """
        Initialize the FolderDataset.

        Args:
            path (str): Path to the folder containing images.
            gpus (str): GPU(s) to use for processing.
        """
        super(Dataset, self).__init__()
        file_types = ["jpg", "png", "dcm"]
        self.fileloader = FileLoader("")
        self.files = [
            os.path.join(path, i)
            for i in os.listdir(path)
            if i.split(".")[-1].lower() in file_types
        ]

    def collate_fn(self, batch):
        """
        Custom collate function for batching.

        Args:
            batch: List of samples to batch.

        Returns:
            dict: Batched data.
        """
        out_dict = {}
        for b in batch:
            for k in b.keys():
                if k in out_dict.keys():
                    out_dict[k] += [b[k]]
                else:
                    out_dict[k] = [b[k]]

        for k in out_dict.keys():
            if k == "data":
                out_dict[k] = torch.cat(out_dict[k], dim=0)

        return out_dict

    def __getitem__(self, index):
        """
        Get a sample from the dataset.

        Args:
            index (int): Index of the sample.

        Returns:
            dict: Sample data.
        """
        return self.fileloader.load_file(self.files[index])

    def __len__(
        self,
    ):
        """
        Get the length of the dataset.

        Returns:
            int: Length of the dataset.
        """
        return len(self.files)



def get_folder_loader(
    path: str, gpus: str, batch_size: int
) -> torch.utils.data.DataLoader:
    """
    Get DataLoader for a folder dataset.

    Args:
        path (str): Path to the folder containing images.
        gpus (str): GPU(s) to use for processing.
        batch_size (int): Batch size.

    Returns:
        torch.utils.data.DataLoader: DataLoader for the folder dataset.
    """
    dataset = FolderDataset(path, gpus)
    loader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=4,
        collate_fn=dataset.collate_fn,
    )
    return loader

class FileLoader:
    """
    Class to load files (images or DICOM) from disk.
    """

    def __init__(self, gpus: str):
        """
        Initialize the FileLoader.

        Args:
            gpus (str): GPU(s) to use for processing.
        """
        self.base_size = 512
        self.file_types = {
            "jpg": self.load_image,
            "png": self.load_image,
            "dcm": self.load_dicom,
        }

        if "cpu" in gpus:
            self.gpus = "cpu"
        else:
            self.gpus = [int(i) for i in gpus.split(",") if len(i) > 0]

    def to_gpu(self, array: torch.tensor) -> torch.tensor:
        """
        Move tensor to GPU.

        Args:
            array (torch.tensor): Input tensor.

        Returns:
            torch.tensor: Tensor moved to GPU if available.
        """
        if len(self.gpus) > 0 and self.gpus != "cpu":
            assert torch.cuda.is_available()
            return array.to(torch.device("cuda:{}".format(self.gpus[0])))
        else:
            return array

    def normalize(self, array: torch.tensor) -> torch.tensor:
        """
        Normalize image tensor.

        Args:
            array (torch.tensor): Input tensor.

        Returns:
            torch.tensor: Normalized tensor.
        """
        assert array.shape[0] == 3, f"{array.shape}"
        # ImageNet normalization
        # Array to be assumed in range [0,1]
        assert (array.min() >= 0) and (array.max() <= 1)

        array = (array - torch.tensor([0.485, 0.456, 0.406]).view(-1, 1, 1)) / (
            torch.tensor([0.229, 0.224, 0.225]).view(-1, 1, 1)
        )
        return array

    def load_file(self, file_path: str) -> dict:
        """
        Load file based on its extension.

        Args:
            file_path (str): Path to the file.

        Returns:
            dict: Loaded file data.
        """
        assert file_path.split(".")[-1].lower() in list(
            self.file_types.keys()
        ), f"filetype not supported: {file_path.split('.')[-1].lower()}"

        return self.file_types[file_path.split(".")[-1].lower()](file_path)

    def load_image(self, image_path: str) -> dict:
        """
        Load image from file.

        Args:
            image_path (str): Path to the image file.

        Returns:
            dict: Image data.
        """
        array = np.array(Image.open(image_path).convert(mode="RGB"))
        array = np.transpose(array, [2, 0, 1])
        original_array = np.copy(array)
        orig_file_size = array.shape[-2:]
        array = torch.tensor(array).float() / 255
        array = self.normalize(array)
        array = self.to_gpu(array)
        return {
            "data": F.interpolate(array.unsqueeze(0), self.base_size),
            "orig_data": original_array,
            "filename": image_path,
            "file_size": orig_file_size,
        }

    def load_dicom(self, image_path: str) -> dict:
        """
        Load DICOM image.

        Args:
            image_path (str): Path to the DICOM file.

        Returns:
            dict: DICOM image data.
        """
        image = sitk.ReadImage(image_path)
        array_view = sitk.GetArrayFromImage(image).astype(np.float32)
        assert (len(array_view.shape) == 3) and (array_view.shape[0] == 1)
        orig_file_size = array_view.shape[-2:]
        tensor = torch.tensor(array_view).float()
        tensor = (tensor - tensor.min()) / (tensor.max() - tensor.min())
        tensor = torch.cat([tensor, tensor, tensor], 0)
        original_array = (tensor.numpy() * 255).astype(np.uint8)
        array = self.normalize(tensor)
        array = self.to_gpu(array)
        return {
            "data": F.interpolate(array.unsqueeze(0), self.base_size),
            "orig_data": original_array,
            "filename": image_path,
            "file_size": orig_file_size,
        }


class FileSaver:
    """
    Class to save prediction results.
    """

    def __init__(
        self,
    ):
        """
        Initialize the FileSaver.
        """
        self.save_modes = {
            "npz": self.export_prediction_as_npz,
            "npy": self.export_prediction_as_numpy,
            "jpg": self.export_prediction_as_jpg,
            "png": self.export_prediction_as_png,
            "json": self.export_prediction_as_json,
            "dicom-seg": self.export_prediction_as_dicomseg,
        }

    def save_prediction(
        self, mask: np.array, outdir: str, file_name: str, mode: str
    ) -> None:
        """
        Save prediction results.

        Args:
            mask (np.array): Prediction mask.
            outdir (str): Output directory.
            file_name (str): File name.
            mode (str): Save mode.
        """
        assert mode in list(self.save_modes.keys())
        assert len(mask.shape) == 3
        self.save_modes[mode](mask, outdir, file_name)

    def export_prediction_as_dicomseg(
        self, mask: np.array, outdir: str, file_name: str
    ) -> None:
        """
        Export prediction as DICOM segmentation.

        Args:
            mask (np.array): Prediction mask.
            outdir (str): Output directory.
            file_name (str): File name.
        """
        assert (
            os.path.splitext(file_name)[-1] == ".dcm"
        ), "Input file has to be dicom to be stored \
            as dicom-seg, it was --{}--".format(
            os.path.splitext(file_name)[-1]
        )
        out_dir = os.path.join(outdir, os.path.splitext(file_name)[0].split("/")[-1])
        os.makedirs(out_dir, exist_ok=True)

        write_dicom_seg(
            metainfo=os.path.join(str(this_directory), "data/metainfo.json"),
            dcm_file=file_name,
            mask=mask.astype(np.uint8),
            out_dir=out_dir,
            id_label_dict=id2label_dict,
        )

    def export_prediction_as_jpg(
        self, mask: np.array, outdir: str, file_name: str
    ) -> None:
        """
        Export prediction as JPG images.

        Args:
            mask (np.array): Prediction mask.
            outdir (str): Output directory.
            file_name (str): File name.
        """
        assert len(mask.shape) == 3
        fileroot = os.path.splitext(file_name)[0].split("/")[-1]
        outdir = os.path.join(outdir, fileroot)
        os.makedirs(outdir, exist_ok=True)

        for i in range(len(mask)):
            out_path = os.path.join(outdir, id2label_dict[str(i)] + ".jpg")
            Image.fromarray(mask[i]).convert("1").save(out_path)

    def export_prediction_as_png(
        self, mask: np.array, outdir: str, file_name: str
    ) -> None:
        """
        Export prediction as PNG images.

        Args:
            mask (np.array): Prediction mask.
            outdir (str): Output directory.
            file_name (str): File name.
        """
        assert len(mask.shape) == 3
        fileroot = os.path.splitext(file_name)[0].split("/")[-1]
        outdir = os.path.join(outdir, fileroot)
        os.makedirs(outdir, exist_ok=True)

        for i in range(len(mask)):
            out_path = os.path.join(outdir, id2label_dict[str(i)] + ".png")
            Image.fromarray(mask[i]).convert("1").save(out_path)

    def export_prediction_as_numpy(
        self, mask: np.array, outdir: str, file_name: str
    ) -> None:
        """
        Export prediction as NumPy arrays.

        Args:
            mask (np.array): Prediction mask.
            outdir (str): Output directory.
            file_name (str): File name.
        """
        os.makedirs(outdir, exist_ok=True)
        out_path = os.path.join(
            outdir, os.path.splitext(file_name)[0].split("/")[-1] + ".npy"
        )
        np.save(out_path, mask)

    def export_prediction_as_npz(
        self, mask: np.array, outdir: str, file_name: str
    ) -> None:
        """
        Export prediction as compressed NumPy arrays.

        Args:
            mask (np.array): Prediction mask.
            outdir (str): Output directory.
            file_name (str): File name.
        """
        os.makedirs(outdir, exist_ok=True)
        out_path = os.path.join(
            outdir, os.path.splitext(file_name)[0].split("/")[-1] + ".npz"
        )
        np.savez_compressed(out_path, mask)

    def export_prediction_as_json(
        self,
        mask: np.array,
        outdir: str,
        file_name: str,
        img_id: int = 1,
        base_ann_id: int = 1,
    ) -> None:
        """
        Export prediction as JSON in COCO format.

        Args:
            mask (np.array): Prediction mask.
            outdir (str): Output directory.
            file_name (str): File name.
            img_id (int, optional): Image ID. Defaults to 1.
            base_ann_id (int, optional): Base annotation ID. Defaults to 1.
        """
        coco_format = get_coco_json_format()
        coco_format["categories"] = create_category_annotation(category_ids)

        coco_format["images"] = []
        coco_format["images"].append([{"id": img_id, "file_name": file_name}])

        coco_format["annotations"] = mask_to_annotation(
            mask=mask, base_ann_id=base_ann_id, img_id=img_id
        )

        os.makedirs(outdir, exist_ok=True)
        out_path = os.path.join(
            outdir, os.path.splitext(file_name)[0].split("/")[-1] + ".json"
        )

        with open(out_path, "w") as outfile:
            json.dump(coco_format, outfile)



import json, os
import colorcet as cc
from pathlib import Path

this_directory = Path(__file__).parent
paxray_labels = json.load(open(os.path.join(str(this_directory), 'data/paxray_labels.json')))
id2label_dict = paxray_labels['label_dict']

thoracic_spine = ["thoracic spine",
                         "vertebrae T1",
                        "vertebrae T2",
                        "vertebrae T3",
                        "vertebrae T4",
                        "vertebrae T5",
                        "vertebrae T6",
                        "vertebrae T7",
                        "vertebrae T8",
                        "vertebrae T9",
                        "vertebrae T10",
                        "vertebrae T11",
                        "vertebrae T12",
                         ]

all_vertebrae = [
    "vertebrae L1",
    "vertebrae L2",
    "vertebrae L3",
    "vertebrae L4",
    "vertebrae L5",
    "vertebrae T1",
    "vertebrae T2",
    "vertebrae T3",
    "vertebrae T4",
    "vertebrae T5",
    "vertebrae T6",
    "vertebrae T7",
    "vertebrae T8",
    "vertebrae T9",
    "vertebrae T10",
    "vertebrae T11",
    "vertebrae T12",
    "vertebrae C1",
    "vertebrae C2",
    "vertebrae C3",
    "vertebrae C4",
    "vertebrae C5",
    "vertebrae C6",
        ]

cervical_spine = [
    "vertebrae C1",
    "vertebrae C2",
    "vertebrae C3",
    "vertebrae C4",
    "vertebrae C5",
    "vertebrae C6",
]

lumbar_spine = [
    "lumbar spine",
    "vertebrae L1",
    "vertebrae L2",
    "vertebrae L3",
    "vertebrae L4",
    "vertebrae L5",
    ]

clavicle_set = ['clavicles',  "clavicle left", "clavicle right"]

scapula_set = [
"scapulas",
"scapula left",
"scapula right",
]

humerus = [
"humerus",
"humerus left",
"humerus right",
    ]

rib = [
    "posterior 12th rib right",
    "posterior 12th rib left",
    "anterior 11th rib right",
    "posterior 11th rib right",
    "anterior 11th rib left",
    "posterior 11th rib left",
    "anterior 10th rib right",
    "posterior 10th rib right",
    "anterior 10th rib left",
    "posterior 10th rib left",
    "anterior 9th rib right",
    "posterior 9th rib right",
    "anterior 9th rib left",
    "posterior 9th rib left",
    "anterior 8th rib right",
    "posterior 8th rib right",
    "anterior 8th rib left",
    "posterior 8th rib left",
    "anterior 7th rib right",
    "posterior 7th rib right",
    "anterior 7th rib left",
    "posterior 7th rib left",
    "anterior 6th rib right",
    "posterior 6th rib right",
    "anterior 6th rib left",
    "posterior 6th rib left",
    "anterior 5th rib right",
    "posterior 5th rib right",
    "anterior 5th rib left",
    "posterior 5th rib left",
    "anterior 4th rib right",
    "posterior 4th rib right",
    "anterior 4th rib left",
    "posterior 4th rib left",
    "anterior 3rd rib right",
    "posterior 3rd rib right",
    "anterior 3rd rib left",
    "posterior 3rd rib left",
    "anterior 2nd rib right",
    "posterior 2nd rib right",
    "anterior 2nd rib left",
    "posterior 2nd rib left",
    "anterior 1st rib right",
    "posterior 1st rib right",
    "anterior 1st rib left",
    "posterior 1st rib left",
]

ribsuper = [
    "12th rib",
    "12th rib",
    "anterior 11th rib",
    "posterior 11th rib",


    "anterior 10th rib",
    "posterior 10th rib",


    "anterior 9th rib",
    "posterior 9th rib",
    
    "anterior 8th rib",
    "posterior 8th rib",


    "anterior 7th rib",
    "posterior 7th rib",


    "anterior 6th rib",
    "posterior 6th rib",


    "anterior 5th rib",
    "posterior 5th rib",
    
    "anterior 4th rib",
    "posterior 4th rib",

    "anterior 3rd rib",
    "posterior 3rd rib",

    "anterior 2nd rib",
    "posterior 2nd rib",

    "anterior 1st rib",
    "posterior 1st rib",
]

diaphragm = ["diaphragm","left hemidiaphragm","right hemidiaphragm",]

mediastinum = [
    "cardiomediastinum",
    "upper mediastinum",
    "lower mediastinum",
    "anterior mediastinum",
    "middle mediastinum",
    "posterior mediastinum",
]

abdomen = [
"stomach", 
"small bowel", 
"duodenum", 
# "gallbladder", 
"liver", 
"pancreas", 
"kidney left", 
"kidney right", 
]

heart = [
"heart",
"heart atrium left",
"heart atrium right",
"heart myocardium",
"heart ventricle left",
"heart ventricle right",
]

breast = [
# "breast",
"breast left",
"breast right",
]

trachea = [
"trachea",
"tracheal bifurcation",
]

zones = [
 "right upper zone lung", 
"right mid zone lung", 
"right lung base", 
"right apical zone lung", 
"left upper zone lung", 
"left mid zone lung", 
"left lung base", 
"left apical zone lung"   , 
]

lung_halves = [
 "right lung", 
    "left lung"   , 
]
vessels = [
"heart",


"ascending aorta",
"descending aorta",
"aortic arch",

"pulmonary artery",
"inferior vena cava",
]

lobes = [
"lung lower lobe left",
"lung upper lobe left",
"lung lower lobe right",
"lung middle lobe right",
"lung upper lobe right",
]



label_mapper = {id2label_dict[k]:[int(k)] for k in id2label_dict.keys()}

label_mapper = {
    **label_mapper, 
    **{
    'thoracic spine': [label_mapper[i] for i in thoracic_spine],
    'all vertebrae': [label_mapper[i] for i in all_vertebrae],
    'cervical spine': [label_mapper[i] for i in cervical_spine],
    'lumbar spine': [label_mapper[i] for i in lumbar_spine],
    'clavicle set': [label_mapper[i] for i in clavicle_set],
    'scapula set': [label_mapper[i] for i in scapula_set],
    'ribs': [label_mapper[i] for i in rib],
    'ribs super': [label_mapper[i] for i in ribsuper],
    'diaphragm': [label_mapper[i] for i in diaphragm],
    'mediastinum': [label_mapper[i] for i in mediastinum],
    'abdomen': [label_mapper[i] for i in abdomen],
    'heart region': [label_mapper[i] for i in heart],
    'breast tissue': [label_mapper[i] for i in breast],
    'trachea': [label_mapper[i] for i in trachea],
    'lung zones': [label_mapper[i] for i in zones],
    'lung halves': [label_mapper[i] for i in lung_halves],
    'vessels': [label_mapper[i] for i in vessels],
    'lung lobes': [label_mapper[i] for i in lobes],
    }
}

colors = [cc.cm.glasbey_bw_minc_20(i) for i in range(len(id2label_dict.keys()))]
colors_alpha = [[i[0],i[1],i[2],i[3]/2] for i in colors]
category_colors = {colors[i][:3]: i for i in range(len(colors))}
category_ids = {id2label_dict[k]:int(k) for k in id2label_dict.keys()}




import sys,os
from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "docs/shortREADME.rst").read_text()

setup(name='cxas',
        version='0.0.11',
        description='Segmentation of 159 anatomical classes for Chest X-Rays.',
        long_description = long_description,
        long_description_content_type="text/x-rst",
        url='https://github.com/ConstantinSeibold/ChestXRayAnatomySegmentation',
        author='Constantin Seibold',
        author_email='constantin.seibold2@uk-essen.de',
        python_requires='>=3.9',
        packages=['cxas', 'cxas.io_utils', 'cxas.models', 'cxas.extraction', 'cxas.models.UNet'],
        package_data={'cxas':['data/*.json']},
        install_requires=[
            'cython',
            'torch',
            'gdown',
            'torchvision',
            'numpy',
            'SimpleITK',
            'pydicom',
            'pydicom_seg',
            'scikit-image',
            'scikit-learn',
            'opencv-python',
            'colorcet',
            'pycocotools',
            'pandas',
            'tqdm',
        ],
        zip_safe=False,
        keywords="chest x-ray anatomy segmntation pytorch",
        classifiers=[
            'Development Status :: 3 - Alpha', 
            'Intended Audience :: Science/Research',
            'Programming Language :: Python',
            'Topic :: Scientific/Engineering',
            'Operating System :: Unix',    
             "License :: Free for non-commercial use",
        ],
        scripts=[
            'bin/cxas', 'bin/cxas_feat_extract', 'bin/cxas_segment',
        ]
    )